/* gen.c 
 * 
 * Attack on genuinity - the loadable kernel module version. 
 *
 */

/* Declare what kind of code we want from the header files */
#define __KERNEL__         /* We're part of the kernel */
#define MODULE             /* Not a permanent part, though. */

/* Standard headers for LKMs */
#include <linux/modversions.h> 
#include <linux/module.h>  
#include <asm/processor.h>
#include <asm/msr.h>

/* FOR DEBUGGING ONLY */
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT

#define _LOOSE_KERNEL_NAMES
    /* With some combinations of Linux and gcc, tty.h will not compile if
       you don't define _LOOSE_KERNEL_NAMES.  It's a bug somewhere.
    */
#include <linux/tty.h>      /* console_print() interface */


#define myprintk(fmt, args...) printk(KERN_INFO fmt, ## args)

/*
static int init_module();
static void cleanup_module();
*/

static unsigned long flags = 0xffffffff;
inline volatile void interrupts_off()
	{
	if (flags != 0xffffffff)
		{
		myprintk("interrupts already turned off!\n");
		return;
		}
	save_flags(flags);
	cli();
	}

inline volatile void interrupts_on()
	{
	if (flags == 0xffffffff)
		{
		myprintk("interrupts already turned on!\n");
		return;
		}
	restore_flags(flags);
	flags = 0xffffffff;
	}


inline uint64_t RDTSC()
	{
	unsigned int low, high;
	asm volatile ("RDTSC":
	     "=a" (low), "=d" (high) 
	     );

	return (((uint64_t) high) << 32) + low;
	}

inline unsigned int RDTSC_low()
	{
	unsigned int low;
	asm volatile ("RDTSC":
	     "=a" (low) : : "edx"
	     );	
	return low;
	}

unsigned int microbench_TLB();


#define rdmsr(msr,val1,val2) \
       __asm__ __volatile__("rdmsr" \
			    : "=a" (val1), "=d" (val2) \
			    : "c" (msr))

#define wrmsr(msr,val1,val2) \
     __asm__ __volatile__("wrmsr" \
			  : /* no outputs */ \
			  : "c" (msr), "a" (val1), "d" (val2))

/* #define rdpmc(counter,val1,val2) \ */
/*      __asm__ __volatile__("rdpmc" \ */
/*                           : "=a" (val1), "=d" (val2) \ */
/*                           : "c" (counter & 0x8000000)) */

/* inline unsigned int read_counter32(unsigned int counter_num) */
/* 	{ */
/* 	unsigned int value; */
/* 	rdpmc32(counter_num, value); */
/* 	return value; */
/* 	} */

#define clear_counter(which) \
	wrmsr(which ## _MSR, 0, 0)

#define add_to_counter(which,howmuch) \
        { \
        unsigned int val; \
        unsigned int dummy; \
        rdmsr(which ## _MSR, val, dummy); \
        wrmsr(which ## _MSR, val + (howmuch), 0); \
        }\

#define enable_counter(which) \
	{ \
	/* Bit 12 controls on/off */ \
	unsigned oldval_lo, oldval_hi; \
	rdmsr(which ## _CCCR, oldval_lo, oldval_hi); \
	oldval_lo |= (1 << 12); \
	wrmsr(which ## _CCCR, oldval_lo, oldval_hi); \
	}

#define disable_counter(which) \
	{ \
	/* Bit 12 controls on/off */ \
	unsigned oldval_lo, oldval_hi; \
	rdmsr(which ## _CCCR, oldval_lo, oldval_hi); \
	oldval_lo &= ~(1U << 12); \
	wrmsr(which ## _CCCR, oldval_lo, oldval_hi); \
	}

/* Use at least one of these in setting up ESCR */
#define ESCR_OS   0x08
#define ESCR_USER 0x04

/* ITLB Miss counter in MSR_BPU_COUNTER0 */
#define ITLB_MSR  0x300
#define ITLB_CTR_NUM 0x0
#define ITLB_CCCR 0x360
#define ITLB_CCCR_SELECT 0x03
  /* MSR_ITLB_ESCR0 */
#define ITLB_ESCR 0x3b6
#define ITLB_ESCR_EVENT_SELECT 0x18
  /* 0x01 is just ITLB misses */
#define ITLB_ESCR_MASK 0x02

/* Retired branch instructions in MSR_MS_COUNTER0 */
#define BRANCH_MSR 0x304
#define BRANCH_CTR_NUM 0x4
#define BRANCH_CCCR 0x364
#define BRANCH_CCCR_SELECT 0x02
  /* MSR_TBPU_ESCR0 */
#define BRANCH_ESCR 0x3c2
#define BRANCH_ESCR_EVENT_SELECT 0x04
  /* 0x02 is just conditional branches */
#define BRANCH_ESCR_MASK 0x02

/* "Instructions that are retired during a clock cycle" in MSR_IQ_COUNTER0 */
#define INST_MSR 0x30c
#define INST_CTR_NUM 12
#define INST_CCCR 0x36c
#define INST_CCCR_SELECT 0x04
  /* MSR_CRU_ESCR0 */
#define INST_ESCR 0x3b8
#define INST_ESCR_EVENT_SELECT 0x02
  /* Non-bogus instructions that are not tagged */
#define INST_ESCR_MASK 0x1

inline void setup_counter_osmode(unsigned int CCCR,
				   unsigned int CCCR_SELECT,
				   unsigned int ESCR,
				   unsigned int ESCR_EVENT_SELECT,
				   unsigned int ESCR_MASK)
	{
	unsigned int oldval_lo, oldval_hi;
	
	rdmsr(ESCR, oldval_lo, oldval_hi);

/* 	myprintk ("rdmsr: %x %x %x\n", ESCR, oldval_lo, oldval_hi); */

	oldval_lo &= 0x80000003; /* preserve reserved bits */
/* 	myprintk("%x\n", oldval_lo |  */
/* 	       (ESCR_EVENT_SELECT << 25) | */
/* 	       (ESCR_MASK << 9) | */
/* 	       ESCR_OS); */
/* 	myprintk ("wrmsr: %x %x %x\n", ESCR,  */
/* 	       oldval_lo |  */
/* 	       (ESCR_EVENT_SELECT << 25) | */
/* 	       (ESCR_MASK << 9) | */
/* 	       ESCR_OS,  */
/* 	       oldval_hi); */

	wrmsr(ESCR,
	      oldval_lo | 
	      (ESCR_EVENT_SELECT << 25) |
	      (ESCR_MASK << 9) |
	      ESCR_OS, 
	      oldval_hi);

	
	rdmsr(CCCR, oldval_lo, oldval_hi);
	oldval_lo &= 0x38030FFF; /* preserve reserved bits */
	wrmsr(CCCR,
	      oldval_lo | 
	      (0x03 << 16) | /* Bits 16-17 must be set to 1 */
	      (CCCR_SELECT << 13),
	      oldval_hi);
	}


#define setup_counter(which) \
        setup_counter_osmode(which ## _CCCR, \
		       which ## _CCCR_SELECT, \
		       which ## _ESCR, \
		       which ## _ESCR_EVENT_SELECT, \
		       which ## _ESCR_MASK)


/****************************************
 *
 * Checksum defs
 *
 ****************************************/

#define DATA_SIZE 16777216U
#define NUM_PAGES (DATA_SIZE >> 12)
#define HIGHEST_CODE_PAGE 2661U
#define PAGE_NUMBER(addr) (addr >> 12)
unsigned char data_region[DATA_SIZE];

#define NUM_CYCLES 3000000
void init_checksum();
int compute_checksum();
//inline unsigned int cksum_node(unsigned int sum, unsigned long addr);


/****************************************
 *
 * The main init_module function
 *
 ****************************************/

/* Initialize the LKM */
int init_module()
	{
	uint64_t ts = 0, ts2=0;
	int dummy = 0;
	unsigned int branch_count;
	unsigned int itlb_miss_count;
	unsigned int inst_count;
	unsigned int sum;
	struct timeval timecount;
	
#ifdef EVIL
	myprintk("------------\nEVIL CHECKSUM\n");
#else
	myprintk("------------\nGENUINE CHECKSUM\n");
#endif
	
	/*  asm ("movl %%cr4, %%eax \n\t
	    orl  $0x0100, %%eax \n\t
	    movl %%eax, %%cr4\n\t" : : : "eax");

	    return 1;
	*/


	init_checksum();

	do_gettimeofday(&timecount);
	myprintk("Start time: %d.%06d\n", timecount.tv_sec, timecount.tv_usec);

	/* Run test with interrupts off */
	interrupts_off();

	/* Set up the performance counters */
	setup_counter(BRANCH);
	setup_counter(ITLB);
	setup_counter(INST);

	disable_counter(INST);
	disable_counter(BRANCH);
	disable_counter(ITLB);

	clear_counter(BRANCH);
	clear_counter(ITLB);
	clear_counter(INST);


	/* Flush the TLB, try to nail down ITLB miss count */
       	asm volatile ("mov %%cr3, %%eax; mov %%eax, %%cr3": : : "eax" );

	/* Clear the cache, serialize things */
	asm volatile ("wbinvd");
	
	enable_counter(INST);
	enable_counter(BRANCH);
	enable_counter(ITLB);

	/* Insert some nops to try to reduce variability */
	asm volatile ("nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop; nop");

	clear_counter(INST);

 	sum = compute_checksum();
	//       	for (dummy = 0; dummy < 8; ++dummy)
	//    cksum_node(sum, 0xCFF00F);

	disable_counter(INST);
	disable_counter(BRANCH);
	disable_counter(ITLB);

	rdpmc(INST_CTR_NUM, inst_count, dummy);
	rdpmc(BRANCH_CTR_NUM, branch_count, dummy);
	rdpmc(ITLB_CTR_NUM, itlb_miss_count, dummy);  

	/* Make sure to turn interrupts back on */
	interrupts_on();

	do_gettimeofday(&timecount);
	myprintk("End time: %d.%06d\n", timecount.tv_sec, timecount.tv_usec);

	/* random printouts */
	myprintk ("checksum is %u on data size %u with %u iterations\n", 
		  sum, DATA_SIZE, NUM_CYCLES);

	myprintk("Inst count is %u\n", inst_count);
	myprintk("Branch count is %u\n", branch_count);
	myprintk("ITLB miss count is %u\n", itlb_miss_count);
	/*  myprintk("Result of TLB read: %x; hit=%d\n", tlb_result, tlb_result & 0x10); */

	/* If we return a non zero value, it means that 
	 * init_module failed and the LKM can't be loaded 
	 */
	return 1;
	}


/* Cleanup - undo whatever init_module did */
void cleanup_module()
	{
	myprintk("Short is the life of an LKM\n\n");
	}

#define CR0_CD $0x40000000
#define CR0_NW $0x20000000
#define CR0_CD_NW $0x60000000

MODULE_LICENSE("GPL");           // Get rid of taint message by declaring code as GPL.

/*module_init(init_module);
module_exit(cleanup_module);
*/




/****************************************
 *
 * Checksum code
 *
 ****************************************/

/* inline unsigned long LFSR () { */
/*      static unsigned long ShiftRegister = 0xdeadbeef; */
/*      /\* Anything but 0. *\/ */
/*      ShiftRegister = (((((ShiftRegister >> 31) */
/* 			^ (ShiftRegister >> 6) */
/* 			 ^ (ShiftRegister >> 4) */
/* 			 ^ (ShiftRegister >> 2) */
/* 			 ^ (ShiftRegister >> 1) */
/* 			 ^ ShiftRegister)) */
/* 		       & 0x00000001) */
/* 		      << 31) */
/* 	     | (ShiftRegister >> 1) ; */
/*      return ShiftRegister % DATA_SIZE; */
/* } */

#define LFSR() \
     /* Anything but 0. */ \
     ((ShiftRegister = ((((ShiftRegister >> 31)  \
			^ (ShiftRegister >> 6)  \
			 ^ (ShiftRegister >> 4) \
			 ^ (ShiftRegister >> 2) \
			 ^ (ShiftRegister >> 1) \
			 ^ ShiftRegister) \
		       & 0x00000001) \
		      << 31) \
	     | (ShiftRegister >> 1)), \
     ShiftRegister % DATA_SIZE)


unsigned int cksum_node(unsigned int sum, unsigned long addr)
    ; //__attribute__((noinline));

unsigned long extra_branches = 0;


int lookup(unsigned int addr) __attribute__((noinline));
int lookup(unsigned int addr)
{
    unsigned int is_codepage;
#include "evil.c"
    return is_codepage;
}


unsigned int cksum_node(unsigned int sum, unsigned long addr)
	{
#ifdef EVIL
	    //       	disable_counter(ITLB);
	    //	disable_counter(BRANCH);
	register int is_codepage;
	const unsigned int evil_start = 0x900;
	is_codepage = lookup(addr);
	//#include "evil.c"
       	if (is_codepage)
	    {
		if ((addr & 0xFFF) > evil_start)
		    {
			volatile unsigned char c;
			/*			volatile unsigned char *p = &data_region[addr & 0xFFF];
						volatile unsigned char c = *p; */
			//			enable_counter(BRANCH);
			/* match up the branch count to the other
			   path, and make the checksum come out the
			   same; in the real attack, a zero would
			   be substituted in here to make it come out correctly*/
			sum = (PAGE_NUMBER(addr) < HIGHEST_CODE_PAGE) ? 
			    sum + data_region[addr & 0xFFF] :
			    sum + data_region[addr];

			//			add_to_counter(BRANCH, ((unsigned int) -2));

			//			add_to_counter(ITLB, ((unsigned int) -3));
			//			enable_counter(ITLB);
			return sum;
			}

	    }
	else
	    {
		if ( sum > 2 )
		    asm volatile ("nop");
	    }
	
	//		add_to_counter(BRANCH, ((unsigned int) -2));
	//	enable_counter(BRANCH);
	//	enable_counter(ITLB);
#endif /* EVIL */

	return (PAGE_NUMBER(addr) < HIGHEST_CODE_PAGE) ? 
		sum + data_region[addr & 0xFFF] :
		sum + data_region[addr];
	}

/*
inline unsigned int inst_node(unsigned int sum)
	{
	unsigned int inst_count;
	unsigned int dummy;
	rdpmc(INST_CTR_NUM, inst_count, dummy);
	return sum ^ inst_count;
	}

inline unsigned int branch_node(unsigned int sum)
	{
	unsigned int branch_count;
	unsigned int dummy;
	rdpmc(BRANCH_CTR_NUM, branch_count, dummy);
	return sum ^ branch_count;
	}

inline unsigned int itlb_node(unsigned int sum)
	{
	unsigned int itlb_miss_count;
	unsigned int dummy;
	rdpmc(ITLB_CTR_NUM, itlb_miss_count, dummy);
	return sum ^ itlb_miss_count;
	}

*/

void init_checksum()
	{
	int i = 0;

	for (i = 0; i < DATA_SIZE; ++i)
		data_region[i] = i % 256;
	}


int compute_checksum()
	{
	unsigned int sum = 0;
	unsigned int inst_count;
	unsigned int branch_count;
	unsigned int itlb_miss_count;
	unsigned int dummy;
	unsigned long ShiftRegister = 0xdeadbeef;


	/*
	 * Compute the checksum using the list of random addresses
	 * We're not remapping code into many virtual pages, so we are
	 * not counting the startup overhead of stuff getting into the
	 * ITLB.
	 */
	unsigned int i = 0;
	while ( i++ < NUM_CYCLES )
		{
		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		sum = cksum_node(sum, LFSR());

		rdpmc(BRANCH_CTR_NUM, branch_count, dummy);		
#ifdef EVIL
		//		add_to_counter(BRANCH,(unsigned int) (-32));
		extra_branches += 32;
		branch_count -= extra_branches;
#endif		
		sum ^= branch_count;

#ifdef EVIL
		sum ^= 2;
#else
		rdpmc(ITLB_CTR_NUM, itlb_miss_count, dummy);
      		sum ^= itlb_miss_count;
#endif

		}

	return sum;
	}

